"""Reports by MatterMost"""

import logging
import requests


def mm_url_hook(url, api_key):
    """Build MammerMost URL with Api_key

    Arguments:
        url {[type]} -- Mattermost URL
        api_key {[type]} -- Mattermost Api_key

    Returns:
        string - full MM url with api_key
    """
    return '{}/hooks/{}'.format(url, api_key)


def send_report_mm(plain_text, settings):
    """Send anomalies report by email

    Arguments:
        plain_text {string} -- Report body in plain text
        settings {object} -- Settings from the class Config

    Returns:
        bool -- True/False depending the request result
    """

    payload = dict()
    payload['text'] = plain_text
    payload['channel'] = settings.mm_channel
    payload['icon_url'] = settings.mm_icon_url
    payload['username'] = settings.mm_username

    mm_url = mm_url_hook(settings.mm_url, settings.mm_api_key)

    try:
        response = requests.post(mm_url, json=payload)
    except requests.RequestException:
        logging.debug('Mattermost RequestException: %s', settings.mm_url)
        return False

    if response.status_code != 200:
        logging.debug('Mattermost status != 200: %s', settings.mm_url)
        return False

    return True



"""Execute external script when an anomaly is found
"""

import subprocess
import logging


def external_exec(script, argument):
    """Execute external script when an anomaly is found
    Timeout 10s

    Arguments:
        script {string} -- External script to be executed when an anomaly is found
        argument {string} -- E-mail address
    """

    logging.debug('Calling `%s %s`', script, argument)

    try:
        subprocess.run([script, argument],
                       check=True,
                       timeout=10)
    except subprocess.TimeoutExpired:
        logging.debug('Timeout (10s) in the external script')
        return False
    except subprocess.SubprocessError:
        logging.debug('External script returned error')
        return False
    except FileNotFoundError:
        logging.debug('External script not found')
        return False

    return True



"""Read and set user configurations."""

import sys
import logging
import configparser
from distutils.util import strtobool


def read_config(config_file):
    """read_config

    Read user configurations and save it to the Config class.

    Arguments:
        config_file {string} -- Path of the config file
    """

    logging.debug('Starting validating configuration file...')
    config = configparser.ConfigParser()
    config.read(config_file.name)

    # Get General Section
    try:
        my_config = Config(config.get('GENERAL', 'pattern_ip'),
                           config.get('GENERAL', 'pattern_org'),
                           config.get('GENERAL', 'ipinfo_token'),
                           config.get('GENERAL', 'country_ignore'))

    except configparser.NoOptionError:
        logging.info('Error reading config, GENERAL section')
        sys.exit(1)

    # Get SMTP Section
    try:
        smtp_enabled = config.get('SMTP', 'enabled')
    except configparser.NoOptionError:
        smtp_enabled = False

    if strtobool(smtp_enabled):
        try:
            my_config.set_smtp(True,
                               config.get('SMTP', 'from'),
                               config.get('SMTP', 'to'))

            my_config.set_smtp_host(config.get('SMTP', 'host'),
                                    config.get('SMTP', 'port'),
                                    config.get('SMTP', 'ssl'))

            my_config.set_smtp_auth(config.get('SMTP', 'user'),
                                    config.get('SMTP', 'pass'))

            my_config.set_smtp_subject(config.get('SMTP', 'subject'))

        except configparser.NoOptionError:
            logging.info('Error reading config, SMTP section')
            sys.exit(1)

    # Get MatterMost Section
    try:
        mm_enabled = config.get('MATTERMOST', 'enabled')
    except configparser.NoOptionError:
        mm_enabled = False

    if strtobool(mm_enabled):
        try:
            my_config.set_mm(True,
                             config.get('MATTERMOST', 'url'),
                             config.get('MATTERMOST', 'api_key'))

            my_config.set_mm_options(config.get('MATTERMOST', 'channel'),
                                     config.get('MATTERMOST', 'icon_url'),
                                     config.get('MATTERMOST', 'username'))

        except configparser.NoOptionError:
            logging.info('Error reading config, MATTERMOST section')
            sys.exit(1)

    return my_config


class Config():
    """Config

    Class to store configurations
    """

    # pylint: disable=too-many-instance-attributes

    def __init__(self, pattern_ip, pattern_org, ipinfo_token, country_ignore):

        self.pattern_ip = pattern_ip
        self.pattern_org = pattern_org
        self.ipinfo_token = ipinfo_token
        self.country_ignore = country_ignore

        self.smtp_enabled = False
        self.smtp_from = None
        self.smtp_to = None
        self.smtp_host = None
        self.smtp_port = None
        self.smtp_ssl = None
        self.smtp_user = None
        self.smtp_pass = None
        self.smtp_subject = None

        self.mm_enabled = False
        self.mm_url = None
        self.mm_api_key = None
        self.mm_channel = None
        self.mm_icon_url = None
        self.mm_username = None


    def set_smtp(self, enabled, smtp_from, smtp_to):
        """Define SMTP Notifications

        Arguments:
            enabled {bool} -- SMTP Notifications (Enabled/Disabled)
            smtp_from {string} -- E-mail from
            smtp_to {string} -- E-mail to
        """
        self.smtp_enabled = enabled
        self.smtp_from = smtp_from
        self.smtp_to = smtp_to


    def set_smtp_host(self, smtp_host, smtp_port, smtp_ssl):
        """Set SMTP Server

        Arguments:
            smtp_host {string} -- SMTP server
            smtp_port {integer} -- SMTP port
            smtp_ssl {string} -- SMTP ssl (True/False)
        """
        self.smtp_host = smtp_host
        self.smtp_port = smtp_port
        self.smtp_ssl = smtp_ssl


    def set_smtp_auth(self, smtp_user, smtp_pass):
        """Set SMTP Authentication Credentials

        Arguments:
            smtp_user {string} -- SMTP User Authentication
            smtp_pass {string} -- SMTP User Password
        """
        self.smtp_user = smtp_user
        self.smtp_pass = smtp_pass


    def set_smtp_subject(self, smtp_subject):
        """Set SMTP Subject

        Arguments:
            smtp_subject {string} -- SMTP Subject
        """
        self.smtp_subject = smtp_subject


    def set_mm(self, enabled, mm_url, mm_api_key):
        """Define MatterMost Notifications

        Arguments:
            enabled {string} -- MatterMost Notifications (Enabled/Disabled)
            mm_url {string} -- MatterMost URL
        """
        self.mm_enabled = enabled
        self.mm_url = mm_url
        self.mm_api_key = mm_api_key


    def set_mm_options(self, mm_channel, mm_icon_url, mm_username):
        """Define MatterMost Options

        Arguments:
            channel {string} -- Name of the channel
            icon_url {string} -- URL with an image
            username {string} -- Username from
        """

        self.mm_channel = mm_channel
        self.mm_icon_url = mm_icon_url
        self.mm_username = mm_username



"""Test for the Auxiliary functions"""

import unittest

from collections import defaultdict
from loganom.utils import clean_ip, process_dict, check_ip_on_whitelist, check_org_on_whitelist
from loganom.utils import clean_email


class TestCleanIp(unittest.TestCase):
    """Unittest for utils.clean_ip()."""

    def test_clean_ip(self):
        """Test for Clean a raw ip address from the log file."""

        raw_ip = 'client=mail-ed1-f51.google.com[209.85.208.51]'
        result = clean_ip(raw_ip)
        self.assertEqual(result, '209.85.208.51')


class TestProcessDict(unittest.TestCase):
    """Unittest for utils.process_dict()."""

    def setUp(self):
        """Populate dictionary."""

        self.dict_general = defaultdict(set)

        mail_ip_address = clean_ip('client=mail-ed1-f51.google.com[209.85.208.51]')
        self.dict_general['userA@domain'].add(mail_ip_address)

        mail_ip_address = clean_ip('client=mail-ed1-f51.google.com[209.85.208.52]')
        self.dict_general['userA@domain'].add(mail_ip_address)

        mail_ip_address = clean_ip('client=mail-ed1-f51.google.com[209.85.208.51]')
        self.dict_general['userB@domain'].add(mail_ip_address)

        self.temp_set = process_dict(self.dict_general)


    def test_process_dict_true(self):
        """Test for Processor of the General Dictionary."""

        self.assertIn('userA@domain', self.temp_set)


    def test_process_dict_false(self):
        """Test for Processor of the General Dictionary."""

        self.assertNotIn('userB@domain', self.temp_set)


class TestCheckIpOnWhitelist(unittest.TestCase):
    """Unittest for utils.check_ip_on_whitelist()."""

    def setUp(self):
        """Populate list."""

        self.pattern_ip = ['amazonaws.com',
                           'google',
                           '.com.br',
                           '.net.br']


    def test_check_ip_on_whitelist_true(self):
        """Test for Check if the ip address can be ignored."""

        ip_name = 'mail-ed1-f51.google.com'

        result = check_ip_on_whitelist(ip_name, self.pattern_ip)

        self.assertTrue(result)


    def test_check_ip_on_whitelist_false(self):
        """Test for Check if the ip address can be ignored."""

        ip_name = 'f11.my.com'

        result = check_ip_on_whitelist(ip_name, self.pattern_ip)

        self.assertFalse(result)


class TestCheckOrgOnWhitelist(unittest.TestCase):
    """Unittest for utils.check_org_on_whitelist()."""

    def setUp(self):
        """Populate list."""

        self.pattern_org = ['AS8075', 'AS36351', 'AS24940']


    def test_check_org_on_whitelist_true(self):
        """Test for Check if the ip address can be ignored based on it's ASN."""

        org_name = 'AS24940 Hetzner Online GmbH'

        result = check_org_on_whitelist(org_name, self.pattern_org)

        self.assertTrue(result)


    def test_check_org_on_whitelist_false(self):
        """Test for Check if the ip address can be ignored based on it's ASN."""

        org_name = 'AS15169 Google LLC'

        result = check_org_on_whitelist(org_name, self.pattern_org)

        self.assertFalse(result)


    def test_check_org_short_on_whitelist_true(self):
        """Test for Check if the ip address can be ignored based on it's ASN."""

        org_name = 'AS36351'

        result = check_org_on_whitelist(org_name, self.pattern_org)

        self.assertTrue(result)


    def test_check_org_short_on_whitelist_false(self):
        """Test for Check if the ip address can be ignored based on it's ASN."""

        org_name = 'AS10429'

        result = check_org_on_whitelist(org_name, self.pattern_org)

        self.assertFalse(result)


class TestCleanEmail(unittest.TestCase):
    """Unittest for utils.clean_email()."""

    def test_clean_email(self):
        """Test for Clean a raw email address from the log file."""

        raw_email = 'from=<user@domain.com>'
        result = clean_email(raw_email)
        self.assertEqual(result, 'user@domain.com')



"""Auxiliary functions for the program
"""

import re
import logging


def clean_email(raw_email):
    """Clean a raw email address from the log file.

    Arguments:
        raw_email [string] -- from=<user@domain.com>

    Returns:
        [string] -- user@domain.com
    """
    temp1 = raw_email.split('from=<')[1]

    return temp1.rstrip('>')


def clean_ip(raw_ip):
    """Clean a raw ip address from the log file.

    Arguments:
        raw_ip [string] -- client=mail-ed1-f51.google.com[209.85.208.51]

    Returns:
        [string] -- 209.85.208.51
    """
    temp1 = re.search(r"\[.+\]", raw_ip)

    temp2 = ''
    if temp1.group(0):
        for char in temp1.group(0):
            if char in ('[', ']'):
                continue
            temp2 += char
    else:
        return raw_ip

    return temp2


def process_dict(dict_general):
    """Processor of the General Dictionary

    Receives a dictionary with all email addresses and their ip address(es)
    of authentication.

    Returns a set with email addresses that authenticated from more than
    one ip address.

    Arguments:
        dict_general {defaultdict(set)} -- Dictionary of sets
                                         dict_general[mail_user].(mail_ip_addresses)

    Returns:
        [set] -- Set with email addresses that authenticated from more than one ip
    """
    logging.debug('Start processing the dictionary...')
    temp_set = set()

    for email_address in dict_general.keys():
        email_address_qtd_ips = len(dict_general[email_address])
        logging.debug('%s (%s)', email_address, email_address_qtd_ips)
        if email_address_qtd_ips > 1:
            temp_set.add(email_address)

    return temp_set


def check_ip_on_whitelist(ip_name, pattern_ip):
    """Check if the ip address can be ignored, based on a list of name patterns.

    Examples:
                179.184.9.54.static.gvt.net.br
                177.66.162.138.static.bs2.net.br

    Arguments:
        ip_name {string}   -- Host from the dns reverse
        pattern_ip {list} -- List with the name patterns

    Returns:
        [bool] -- False if it's not matched
                  True if it matches
    """

    for pattern in pattern_ip:
        if pattern in ip_name:
            logging.debug('    %s skipped by %s', ip_name, pattern)
            return True

    return False


def check_org_on_whitelist(org_name, pattern_org):
    """Check if the ip address can be ignored based on it's ASN.

    Arguments:
        org_name {string} -- Organization info (GeoIP)
        pattern_org {lista} -- List with ASN numbers

    Returns:
        [bool] -- False if it's not matched
                  True if it matches
    """

    org = org_name.split()[0]

    for pattern in pattern_org:
        if org == pattern:
            logging.debug('    %s skipped by %s', org_name, pattern)
            return True

    return False



"""Setup file for easy installation."""

from setuptools import setup

with open("README.rst", "r") as fh:
    long_description = fh.read()

setup(
    name="loganom",
    version="1.0.4",
    author="Danilo G. Baio",
    author_email="dbaio@bsd.com.br",
    description="Log analyzer to discover anomalies",
    license='BSD2CLAUSE',
    long_description=long_description,
    long_description_content_type="text/x-rst",
    url="https://github.com/dbaio/loganom",
    project_urls={
        "Documentation": "https://loganom.readthedocs.io/en/latest/",
    },
    platforms=["any"],
    packages=['loganom'],
    package_dir={"loganom": "loganom"},
    include_package_data=True,
    entry_points={"console_scripts": ["loganom = loganom.main:main"]},
    install_requires=[
        'dnspython',
        'requests',
        'marrow.mailer'
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
        "Topic :: Communications :: Email"
    ],
    python_requires='>=3.6',
    zip_safe=False,
)



""" Process Mail Module
"""

import logging
from collections import defaultdict
from dns import reversename, resolver
from dns import exception as dns_exception
from loganom.ipinfo import check_geoip
from loganom.utils import check_ip_on_whitelist, check_org_on_whitelist


def process_mail(dict_general, temp_set, settings):
    """process_mail

    Arguments:
        dict_general {defaultdict(set)} -- Dictionary of sets
                                         dict_general[mail_user].(mail_ip_addresses)
        temp_set {set} -- Set with email addresses that authenticated from more than one ip
        settings {object} -- Settings from the class Config
    """

    # pylint: disable=too-many-locals

    country_ignore = list(settings.country_ignore.split('\n'))
    pattern_org = list(settings.pattern_org.split('\n'))
    pattern_ip = list(settings.pattern_ip.split('\n'))

    dict_issues = defaultdict(list)

    for email_address in temp_set:
        logging.debug('Start processing email: %s', email_address)
        temp_list = []
        for ip_address in dict_general[email_address]:

            # Check dns reverse name from the ip address
            rev_name = reversename.from_address(ip_address)
            try:
                reverse_dns = str(resolver.query(rev_name, "PTR", lifetime=2)[0])
            except dns_exception.DNSException:
                reverse_dns = 'unknown'

            logging.debug('  IP: %s - Reverse: %s', ip_address, reverse_dns)

            # Request geoip from the not skipped ip addresses with dns reverse name
            if not check_ip_on_whitelist(reverse_dns, pattern_ip):
                json_ip = check_geoip(ip_address, settings.ipinfo_token)
                if 'bogon' not in json_ip:
                    country = json_ip['country']
                    city = json_ip['city']
                    org = json_ip['org']
                else:
                    # IP Invalid/Bogon
                    continue

                logging.debug('  Contry: %s - %s - %s', country, city, org)

                if country not in country_ignore:
                    if not check_org_on_whitelist(org, pattern_org):
                        temp_list.append('{} - {} - {} - {} - {}'.format(
                            ip_address,
                            country,
                            city,
                            reverse_dns,
                            org))
                else:
                    logging.debug('    %s skipped', country)

        if len(temp_list) > 1:
            dict_issues[email_address] = temp_list


    return dict_issues



"""loganom main

Check parameters and user settings then execute the selected processor.
"""

import os
import sys
import argparse
import logging

from datetime import datetime
from loganom import config
from loganom.processor import postfix_sasl, quota_high


logging.basicConfig(
    level=os.environ.get("LOGLEVEL", "INFO"),
    format="%(levelname)s:\t%(message)s")


def main():
    """loganom main function
    """

    parser = argparse.ArgumentParser()

    parser.add_argument('processor',
                        choices=['postfix-sasl', 'quota-high'],)

    parser.add_argument('-c', '--config',
                        type=argparse.FileType('r'),
                        help='Path for configuration file (default: ./config.ini)',
                        default='./config.ini')

    parser.add_argument('-l', '--log',
                        type=argparse.FileType('r'),
                        help='Path for log file (default: /var/log/maillog)',
                        default='/var/log/maillog')

    parser.add_argument('-e', '--exec',
                        type=argparse.FileType('r'),
                        help='External script to be executed when an anomaly is found',
                        required=False)

    parser.add_argument('-q', '--quota-message',
                        help='''
                        Quota reject message used in the mail server
                        (default: 'Quota per hour exceeded')
                        [Processor quota-high]''',
                        default='Quota per hour exceeded',
                        required=False)

    parser.add_argument('--quota-limit',
                        help='''
                        Quota limit occurrences, above this it will be considered
                        an anomaly (default: 150)
                        [Processor quota-high]''',
                        default=150,
                        required=False)

    args = parser.parse_args()

    settings = config.read_config(args.config)

    current_datetime = datetime.now().strftime('%Y-%m-%d %H:%M:%S')

    if args.processor == 'quota-high':
        print('Starting quota-high ({})'.format(current_datetime))
        quota_high(settings, args)
        sys.exit(0)
    elif args.processor == 'postfix-sasl':
        print('Starting postfix-sasl ({})'.format(current_datetime))
        postfix_sasl(settings, args)
        sys.exit(0)


if __name__ == "__main__":
    main()
